/* 
*	Copyright (C) 2006-2011 Team MediaPortal
*	http://www.team-mediaportal.com
*
*  This Program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2, or (at your option)
*  any later version.
*   
*  This Program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
*  GNU General Public License for more details.
*   
*  You should have received a copy of the GNU General Public License
*  along with GNU Make; see the file COPYING.  If not, write to
*  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
*  http://www.gnu.org/copyleft/gpl.html
*
*/

#include "rtpsource.h"
#include "windows.h"
#include <stdio.h>
#include <shlobj.h>

#ifdef LOGGING
static char logFile[MAX_PATH];
static WORD logFileParsed = -1;

void GetLogFile_(char *pLog)
{
  SYSTEMTIME systemTime;
  GetLocalTime(&systemTime);
  if(logFileParsed != systemTime.wDay)
  {
    TCHAR folder[MAX_PATH];
    ::SHGetSpecialFolderPath(NULL,folder,CSIDL_COMMON_APPDATA,FALSE);
    sprintf_s(logFile,"%s\\Team MediaPortal\\MediaPortal TV Server\\log\\MPIPTVSource-%04.4d-%02.2d-%02.2d.Log",folder, systemTime.wYear, systemTime.wMonth, systemTime.wDay);
    logFileParsed=systemTime.wDay; // rec
  }
  strcpy_s(pLog, MAX_PATH, &logFile[0]);
}

void LogDebug_(const char *fmt, ...)
{
  va_list ap;
  va_start(ap,fmt);

  char buffer[1000];
  int tmp;
  va_start(ap,fmt);
  tmp=vsprintf_s(buffer, fmt, ap);
  va_end(ap);
  SYSTEMTIME systemTime;
  GetLocalTime(&systemTime);

  TCHAR filename[1024];
  GetLogFile_(filename);
  FILE* fp;
  fopen_s(&fp, filename,"a+");

  if (fp!=NULL)
  {
    fprintf(fp,"%02.2d-%02.2d-%04.4d %02.2d:%02.2d:%02.2d.%03.3d [%x]%s\n",
      systemTime.wDay, systemTime.wMonth, systemTime.wYear,
      systemTime.wHour,systemTime.wMinute,systemTime.wSecond,
      systemTime.wMilliseconds,
      GetCurrentThreadId(),
      buffer);
    fclose(fp);
  }
  char buf[1000];
  sprintf_s(buf,"%02.2d-%02.2d-%04.4d %02.2d:%02.2d:%02.2d %s\n",
    systemTime.wDay, systemTime.wMonth, systemTime.wYear,
    systemTime.wHour,systemTime.wMinute,systemTime.wSecond,
    buffer);
  ::OutputDebugString(buf);
};

#endif

RtpSource::RtpSource() 
{
  rtpSsrc = 1;
  isFirstPacket = true;
  rtpStreamRunning = false;
  firstItem = NULL;
  lastItem = NULL;
  packetSizeTotal = 0;
}

RtpSource::~RtpSource() 
{
  // Clean buffer
  while (firstItem != NULL) 
  {
    PacketItem *temp = firstItem;
    firstItem = firstItem->next;
    delete temp->data;
    delete temp;
  }
  firstItem = NULL;
  lastItem = NULL;
}

void RtpSource::AddPacket(char* data, int length, unsigned int timestamp, unsigned int sequenceNr) 
{
#ifdef EXTENSIVELOGGING
  LogDebug_("RTP: Enter addPacket. SequenceNr: %u", sequenceNr);
#endif
  PacketItem* newItem = new PacketItem();
  newItem->data = new char[length];
  memcpy(newItem->data, data, length);
  newItem->size = length;
  newItem->next = NULL;
  newItem->sequenceNr = sequenceNr;
  newItem->timestamp = timestamp;

  // Is the buffer empty?
  if (firstItem == NULL) 
  {
    firstItem = newItem;
    lastItem = newItem;
    packetSizeTotal = length;
    // This is the first item. Set "lastSequenceNumber" accordingly
    this->lastSequenceNr = newItem->sequenceNr == 0 ? 0x0000FFFF : newItem->sequenceNr-1;
#ifdef EXTENSIVELOGGING
    LogDebug_("RTP: Added firstItem");
#endif

  } 
  else 
  {
    // Check if this packet belongs after the last packet (most likely situation, handle this first)
    if ((newItem->sequenceNr > lastItem->sequenceNr) ||
      // The sequence number restarts after 65536, this must be handled
      (lastItem->sequenceNr >= 0xFFF0 && newItem->sequenceNr <= 0x000F)) 
    {
      lastItem->next = newItem;
      packetSizeTotal += length;
      lastItem = newItem;
#ifdef EXTENSIVELOGGING
      //LogDebug_("RTP: Packet added in order");
#endif
      // Check if this packet is far too late (would belong before the first packet in the buffer) and drop it
    } 
    else if (newItem->sequenceNr < firstItem->sequenceNr-1) 
    {
      delete newItem->data;
#ifdef EXTENSIVELOGGING
      LogDebug_("RTP: Packet dropped. SequenceNr: %d", newItem->sequenceNr);
#endif
      delete newItem;

     // The packet belongs somewhere in the middle of the buffer. Find the position and post data in there
    } 
    else 
    {
      PacketItem *currentItem = firstItem;
      while (currentItem != NULL && !PacketBelongsBetween(newItem, currentItem, currentItem->next)) 
      {
        currentItem = currentItem->next;
      }
      // Check if we found the position
      if (currentItem != NULL && currentItem->next != NULL) 
      {
        // Insert new item
        newItem->next = currentItem->next;
        currentItem->next = newItem;
        packetSizeTotal += newItem->size;
        PacketItem *tempItem = firstItem;
#ifdef EXTENSIVELOGGING
      LogDebug_("RTP: Packet added between");
#endif

      } 
      else 
      {
        // This shouldn't happen at all!!
        delete newItem->data;
        delete newItem;
#ifdef LOGGING
        LogDebug_("RTP: ERROR!!!! while adding Packet.");
#endif
      }
    }
  }
}

bool RtpSource::PacketBelongsBetween(PacketItem *item, PacketItem *itemBefore, PacketItem *itemAfter) 
{
#ifdef EXTENSIVELOGGING
  //LogDebug_("RTP: Enter packetBelongsBetween: %d, %d, %d", item->sequenceNr, itemBefore->sequenceNr, itemAfter->sequenceNr);
#endif

  if (item == NULL || itemBefore == NULL || itemAfter == NULL) 
    return false;
  // Most likely situation: the packets are in ascending order 
  if (itemBefore->sequenceNr < itemAfter->sequenceNr) 
  {
    return (item->sequenceNr > itemBefore->sequenceNr && item->sequenceNr < itemAfter->sequenceNr);
  // The sequnce number is resetted between itemBefore and itemAfter
  } 
  else 
  {
    return (item->sequenceNr > itemBefore->sequenceNr || item->sequenceNr < itemAfter->sequenceNr);
  }
}

bool RtpSource::SequenceIsContinuous(unsigned int sequnceNr0, unsigned int sequnceNr1) 
{
#ifdef EXTENSIVELOGGING
  //LogDebug_("RTP: Enter sequenceIsContinous: %d, %d", sequnceNr0, sequnceNr1);
#endif
  return (sequnceNr0+1 == sequnceNr1) || (sequnceNr0 == 0xFFFF && sequnceNr1 == 0);
}

int RtpSource::GetPacketData(char* buffer, int bufferSize, unsigned int &firstTimestamp, unsigned int &lastTimestamp, bool &isContinous) 
{
  int bytesTransferred = 0;
  int firstsequenceNr = 0;

  // Check if the first packet is continous to last one
  isContinous = SequenceIsContinuous(lastSequenceNr, firstItem->sequenceNr);

  // Duplicate first data if there are some packets missing in the beginning
  while (!SequenceIsContinuous(lastSequenceNr, firstItem->sequenceNr)) 
  {
#ifdef EXTENSIVELOGGING
   LogDebug_("RTP: Add dummy packet at beginning. LastSeqNr: %d, new firstSequenceNr: %d", lastSequenceNr, firstItem->sequenceNr);
#endif
    PacketItem* newItem = new PacketItem();
    // Test: Zero memory instead of copying data
    newItem->data = new char[firstItem->size];
    ZeroMemory(newItem->data, firstItem->size);
    newItem->size = firstItem->size;
    newItem->next = firstItem;
    newItem->sequenceNr = firstItem->sequenceNr-1;
    if (newItem->sequenceNr > 0xFFFF) newItem->sequenceNr = 0xFFFF;
    newItem->timestamp = 0;
    firstItem = newItem;
  }

  // Get first item
  PacketItem* currentItem = firstItem;
  // Get data. Stop when there are no more packets or when there is a paket missing in the stream or the output buffer is filled
  while (currentItem != NULL && currentItem->next != NULL && SequenceIsContinuous(lastSequenceNr, currentItem->sequenceNr) && (bytesTransferred + currentItem->size) <= bufferSize) 
  {
    // Copy data to destination buffer
    memcpy(buffer+bytesTransferred, currentItem->data, currentItem->size);

    // First item?
    if (bytesTransferred == 0) 
    {
      firstTimestamp = currentItem->timestamp;
      firstsequenceNr = currentItem->sequenceNr;
#ifdef EXTENSIVELOGGING
      if (!isContinous)
        LogDebug_("RTP: Got unContinous packet. LastSequenceNr: %d, currentSequenceNr: %d", lastSequenceNr, currentItem->sequenceNr);
#endif
    }
    // Set new total copied bytes
    bytesTransferred += currentItem->size;

    // Set last data (this might be the last item to process)
    lastTimestamp = currentItem->timestamp;
    lastSequenceNr = currentItem->sequenceNr;

    packetSizeTotal -= currentItem->size;
        
    // Remember current Item
    PacketItem *temp = currentItem;

    // Set next current item
    currentItem = currentItem->next;

    // Check if next packet is continous. If not, add dummy-data
    // Note: If this is not done, the TV-Server hangs from time to time (picture freezes, but data is still received)
    //       Perhaps when only one packet is processed??
    //       Tested on 9/11/2010 with TV-Server 1.2.0 alpha
    int currentSequenceNr = temp->sequenceNr;
    while (!SequenceIsContinuous(currentSequenceNr, currentItem->sequenceNr) && (bytesTransferred + currentItem->size) <= bufferSize) 
    {
      // Test: Zero memory instead of copying data
      ZeroMemory((buffer+bytesTransferred), currentItem->size);
      bytesTransferred += currentItem->size;
      currentSequenceNr++;
      isContinous = false;
      if (currentSequenceNr > 0xFFFF) currentSequenceNr = 0;
    }
    
    // Free some memory
    delete temp->data;
    delete temp;
  }
    
  // Set new first Item
  if (currentItem == NULL) 
  {
    firstItem = NULL;
    lastItem = NULL;
  } 
  else
    firstItem = currentItem;

  // Return processed bytes
#ifdef EXTENSIVELOGGING
  LogDebug_("RTP: Sending data to next filter: Len: %d, firstTS: %d, lastTS: %d, firstSeqNr: %d, lastSeqNr: %d, continous: %d", bytesTransferred, firstTimestamp, lastTimestamp, firstsequenceNr, lastSequenceNr, isContinous);
#endif
  return bytesTransferred;
}

bool RtpSource::CheckDataCanBePosted(int bufferSize) 
{
  // If buffer is greater than IPTV_RTP_INTERNAL_BUFFER_SIZE deliver data
#ifdef EXTENSIVELOGGING
  LogDebug_("RTP: Enter CheckDataCanBePosted: %d", bufferSize);
#endif

  if (packetSizeTotal >= IPTV_RTP_INTERNAL_BUFFER_SIZE) 
    return true;

  return false;
}

bool RtpSource::IsRTPpacket(char *data, int datalen)
 {
  if (datalen < RTP_HEADER_SIZE) 
    return false;
     
  // 1. check: First bit must be one, and second bit must be zero
  bool failPacket = ((((char)*(data))&0xC0) != 0x80);
  // 2. check: SSRC (source identifier) must be the same for every packet
  unsigned int ssrc0 = ((unsigned int)*(data+8))&0x000000ff;
  unsigned int ssrc1 = ((unsigned int)*(data+9))&0x000000ff;
  unsigned int ssrc2 = ((unsigned int)*(data+10))&0x000000ff;
  unsigned int ssrc3 = ((unsigned int)*(data+11))&0x000000ff;
  unsigned int ssrc = (ssrc0<<24) + (ssrc1<<16) + (ssrc2<<8) + (ssrc3<<0);
  // Is it the same like last packet?
  if (ssrc != rtpSsrc) 
  {
    // Save ssrc for next packet
    rtpSsrc = ssrc;
    // This is OK on first packet (ssrc unknown at first). On other: This is not a rtp packet
    failPacket &= isFirstPacket;
  }
  isFirstPacket = false;
  return !failPacket;
}

bool RtpSource::ProcessPacket(char *data, int datalen) 
{
#ifdef EXTENSIVELOGGING
      //LogDebug_("RTP: Enter processPacket: %d", datalen);
#endif
  // Padding is used rarely. But we need to handle it
  int padding = 0;
  if (((*data)&0x20) != 0) 
  {
    padding = ((int)*(data+datalen-1))&0x000000ff;
  }

  // Check for header extension
 unsigned int headerExtension = 0;
 if (((*data)&0x10) != 0) 
 {
    unsigned int hx0 = ((unsigned int)*(data+RTP_HEADER_SIZE+2))&0x000000ff;
    unsigned int hx1 = ((unsigned int)*(data+RTP_HEADER_SIZE+3))&0x000000ff;
    headerExtension = (hx0<<8) + (hx1<<0);
    // The header-extension identifier and length field must be taken into account
    headerExtension += 1;
    // Header length is given in 32-bit words
    headerExtension *= 4;
  }

  // The rtp header is 12 bytes long
  if (datalen <= RTP_HEADER_SIZE) 
    return false;
  // Get timestamp
  unsigned int ts0 = ((unsigned int)*(data+4))&0x000000ff;
  unsigned int ts1 = ((unsigned int)*(data+5))&0x000000ff;
  unsigned int ts2 = ((unsigned int)*(data+6))&0x000000ff;
  unsigned int ts3 = ((unsigned int)*(data+7))&0x000000ff;
  unsigned int timestamp = (ts0<<24) + (ts1<<16) + (ts2<<8) + (ts3<<0);

  // Get sequence number
  unsigned int sn0 = ((unsigned int)*(data+2))&0x000000ff;
  unsigned int sn1 = ((unsigned int)*(data+3))&0x000000ff;
  unsigned int sequencenr = (sn0<<8) + (sn1<<0);

  //LogDebug_("RTP: processPacket. SequenceNr: %u", sequencenr);

  // Store packet
  AddPacket(data+RTP_HEADER_SIZE+headerExtension, datalen-RTP_HEADER_SIZE-padding-headerExtension, timestamp, sequencenr);
#ifdef EXTENSIVELOGGING
  //LogDebug_("RTP: Leave ProccesPacket");
#endif
  return true;

}